/*
 Copyright (c) 2011, OpenEmu Team

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
     * Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
     * Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
     * Neither the name of the OpenEmu Team nor the
       names of its contributors may be used to endorse or promote products
       derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY OpenEmu Team ''AS IS'' AND ANY
 EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL OpenEmu Team BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "OELibraryDatabase.h"

#import "OESystemPlugin.h"

#import "NSImage+OEDrawingAdditions.h"

#import "OEDBAllGamesCollection.h"
#import "OEDBSystem.h"
#import "OEDBGame.h"
#import "OEDBRom.h"
#import "OEDBSaveState.h"

#import "OEDBSavedGamesMedia.h"
#import "OEDBScreenshotsMedia.h"
#import "OEDBVideoMedia.h"

#import "OESystemPicker.h"

#import "NSFileManager+OEHashingAdditions.h"

#import "OEFSWatcher.h"
#import "OEROMImporter.h"

#import "NSURL+OELibraryAdditions.h"

#import <OpenEmuBase/OpenEmuBase.h>
#import <OpenEmuSystem/OpenEmuSystem.h>

NSString *const OEDatabasePathKey            = @"databasePath";
NSString *const OEDefaultDatabasePathKey     = @"defaultDatabasePath";
NSString *const OESaveStateLastFSEventIDKey  = @"lastSaveStateEventID";

NSString *const OELibraryDatabaseUserInfoKey = @"OELibraryDatabase";
NSString *const OESaveStateFolderURLKey      = @"saveStateFolder";
NSString *const OEScreenshotFolderURLKey     = @"screenshotFolder";

NSString *const OELibraryRomsFolderURLKey    = @"romsFolderURL";

@interface OELibraryDatabase ()
{
    NSArrayController *_romsController;

    NSManagedObjectModel *_managedObjectModel;
    NSManagedObjectContext *_managedObjectContext;

    NSThread *_syncThread;
}

- (BOOL)loadPersistantStoreWithError:(NSError **)outError;
- (BOOL)loadManagedObjectContextWithError:(NSError **)outError;
- (void)managedObjectContextDidSave:(NSNotification *)notification;

- (NSManagedObjectModel *)managedObjectModel;

- (void)OE_setupStateWatcher;
- (void)OE_removeStateWatcher;

@property(strong) OEFSWatcher *saveStateWatcher;
@property(copy)   NSURL       *databaseURL;
@property(strong) NSPersistentStoreCoordinator *persistentStoreCoordinator;
@end

static OELibraryDatabase *defaultDatabase = nil;

@implementation OELibraryDatabase
@synthesize persistentStoreCoordinator = _persistentStoreCoordinator, databaseURL, importer, saveStateWatcher;

#pragma mark -

+ (BOOL)loadFromURL:(NSURL *)url error:(NSError **)outError
{
    NSLog(@"OELibraryDatabase loadFromURL:%@", url);

    BOOL isDir = NO;
    if(![[NSFileManager defaultManager] fileExistsAtPath:[url path] isDirectory:&isDir] || !isDir)
    {
        if(outError != NULL)
        {
			NSString     *description = NSLocalizedString(@"The OpenEmu Library could not be found.", @"");
			NSDictionary *dict        = [NSDictionary dictionaryWithObject:description forKey:NSLocalizedDescriptionKey];
			
            *outError = [NSError errorWithDomain:@"OELibraryDatabase" code:OELibraryErrorCodeFolderNotFound userInfo:dict];
        }
        return NO;
    }

    defaultDatabase = [[OELibraryDatabase alloc] init];
    [defaultDatabase setDatabaseURL:url];

    if(![defaultDatabase loadPersistantStoreWithError:outError])
    {
        defaultDatabase = nil;
        return NO;
    }

    if(![defaultDatabase loadManagedObjectContextWithError:outError])
    {
        defaultDatabase = nil;
        return NO;
    }

    [[NSUserDefaults standardUserDefaults] setObject:[[defaultDatabase databaseURL] path] forKey:OEDatabasePathKey];
    [defaultDatabase OE_setupStateWatcher];

    dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, (int64_t)(2.0 * NSEC_PER_SEC));
    dispatch_after(popTime, dispatch_get_main_queue(), ^{
        [[defaultDatabase importer] start];
        [defaultDatabase startArchiveVGSync];
    });

    return YES;
}

- (BOOL)loadManagedObjectContextWithError:(NSError **)outError
{
    _managedObjectContext = [[NSManagedObjectContext alloc] initWithConcurrencyType:NSPrivateQueueConcurrencyType];

    NSMergePolicy *policy = [[NSMergePolicy alloc] initWithMergeType:NSMergeByPropertyObjectTrumpMergePolicyType];
    [_managedObjectContext setMergePolicy:policy];

    if(_managedObjectContext == nil) return NO;

    NSPersistentStoreCoordinator *coordinator = [self persistentStoreCoordinator];
    [_managedObjectContext setPersistentStoreCoordinator:coordinator];
    [[_managedObjectContext userInfo] setValue:self forKey:OELibraryDatabaseUserInfoKey];

    // remeber last loc as database path
    NSUserDefaults *standardDefaults = [NSUserDefaults standardUserDefaults];
    [standardDefaults setObject:[self.databaseURL path] forKey:OEDatabasePathKey];

    return YES;
}

- (BOOL)loadPersistantStoreWithError:(NSError **)outError
{
    NSManagedObjectModel *mom = [self managedObjectModel];
    if(mom == nil)
    {
        NSLog(@"%@:%@ No model to generate a store from", [self class], NSStringFromSelector(_cmd));
        return NO;
    }

    NSURL *url = [self.databaseURL URLByAppendingPathComponent:OEDatabaseFileName];
    [self setPersistentStoreCoordinator:[[NSPersistentStoreCoordinator alloc] initWithManagedObjectModel:mom]];

    NSDictionary *options = @{
        NSMigratePersistentStoresAutomaticallyOption : @YES,
        NSInferMappingModelAutomaticallyOption       : @YES,
    };
    if([[self persistentStoreCoordinator] addPersistentStoreWithType:NSSQLiteStoreType configuration:nil URL:url options:options error:outError] == nil)
    {
        [self setPersistentStoreCoordinator:nil];

        return NO;
    }

    DLog(@"ROMS folder url: %@", [self romsFolderURL]);
    return YES;
}

#pragma mark -

+ (OELibraryDatabase *)defaultDatabase
{
    return defaultDatabase;
}

- (id)init
{
    if((self = [super init]))
    {
        _romsController = [[NSArrayController alloc] init];

        OEROMImporter *romImporter = [[OEROMImporter alloc] initWithDatabase:self];
        [romImporter loadQueue];
        [self setImporter:romImporter];

        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(applicationWillTerminate:) name:NSApplicationWillTerminateNotification object:NSApp];
        [[NSUserDefaults standardUserDefaults] addObserver:self forKeyPath:OESaveStateFolderURLKey options:0 context:nil];
    }

    return self;
}

- (void)dealloc
{
    NSLog(@"destroying LibraryDatabase");
    [self OE_removeStateWatcher];
    [[NSUserDefaults standardUserDefaults] removeObserver:self forKeyPath:OESaveStateFolderURLKey];
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}

- (void)awakeFromNib
{
}

- (void)applicationWillTerminate:(id)sender
{
    [self OE_removeStateWatcher];
    [[self importer] saveQueue];

    NSError *error = nil;
    if(![self save:&error])
    {
        [NSApp presentError:error];
        return;
    }

    NSLog(@"Did save Database");
}

- (void)observeValueForKeyPath:(NSString *)keyPath ofObject:(id)object change:(NSDictionary *)change context:(void *)context
{
    NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];

    if(object==defaults && [keyPath isEqualToString:OESaveStateFolderURLKey])
    {
        NSString *path = [[self stateFolderURL] path];

        [self OE_removeStateWatcher];
        [defaults removeObjectForKey:OESaveStateLastFSEventIDKey];
        
        [self OE_setupStateWatcher];
        [[self saveStateWatcher] callbackBlock](path, kFSEventStreamEventFlagItemIsDir);
    }
}
#pragma mark - Administration

- (void)disableSystemsWithoutPlugin
{
    NSArray *allSystems = [OEDBSystem allSystems];
    for(OEDBSystem *aSystem in allSystems)
    {
        if([aSystem plugin]) continue;
        [aSystem setEnabled:[NSNumber numberWithBool:NO]];
    }
}

#pragma mark - Save State Handling

- (void)OE_setupStateWatcher
{
    NSString *stateFolderPath = [[self stateFolderURL] path];
    __block __unsafe_unretained OEFSBlock recFsBlock;
    __block OEFSBlock fsBlock = [^(NSString *path, FSEventStreamEventFlags flags)
    {
      if(flags & kFSEventStreamEventFlagItemIsDir)
        {
            NSFileManager *fileManager = [NSFileManager defaultManager];
            NSURL   *url   = [NSURL fileURLWithPath:path];
            NSError *error = nil;

            if([fileManager fileExistsAtPath:path])
            {
                if([[path pathExtension] isEqualTo:OESaveStateSuffix])
                {
                    dispatch_time_t popTime = dispatch_time(DISPATCH_TIME_NOW, 0.2 * NSEC_PER_SEC);
                    dispatch_after(popTime, dispatch_get_main_queue(), ^{
                        [OEDBSaveState updateOrCreateStateWithPath:path];
                    });
                }
                else
                {
                    NSArray *contents = [fileManager contentsOfDirectoryAtPath:path error:&error];
                    [contents enumerateObjectsUsingBlock:^(NSString *subpath, NSUInteger idx, BOOL *stop) {
                        recFsBlock([path stringByAppendingPathComponent:subpath], flags);
                    }];
                }
            }
            else
            {
                NSFetchRequest *fetchRequest = [NSFetchRequest fetchRequestWithEntityName:@"SaveState"];
                NSPredicate    *predicate    = [NSPredicate predicateWithFormat:@"location BEGINSWITH[cd] %@", [url absoluteString]];
                [fetchRequest setPredicate:predicate];
                NSArray *result = [[self managedObjectContext] executeFetchRequest:fetchRequest error:&error];
                if(error) DLog(@"executing fetch request failed: %@", error);
                [result enumerateObjectsUsingBlock:^(OEDBSaveState *state, NSUInteger idx, BOOL *stop) {
                    [state remove];
                }];
            }
        }
    } copy];

    recFsBlock = fsBlock;

    OEFSWatcher *watcher = [OEFSWatcher persistentWatcherWithKey:OESaveStateLastFSEventIDKey forPath:stateFolderPath withBlock:fsBlock];
    [watcher setDelay:1.0];
    [watcher setStreamFlags:kFSEventStreamCreateFlagUseCFTypes|kFSEventStreamCreateFlagIgnoreSelf|kFSEventStreamCreateFlagFileEvents];

    [self setSaveStateWatcher:watcher];
    [[self saveStateWatcher] startWatching];
}

- (void)OE_removeStateWatcher
{
    NSLog(@"OE_removeStateWatcher");
    [[self saveStateWatcher] stopWatching];
    [self setSaveStateWatcher:nil];
}

#pragma mark - CoreData Stuff

- (NSManagedObjectID *)managedObjectIDForURIRepresentation:(NSURL *)uri
{
    return [[self persistentStoreCoordinator] managedObjectIDForURIRepresentation:uri];
}

- (NSManagedObjectModel *)managedObjectModel
{
    if(_managedObjectModel != nil) return _managedObjectModel;

    NSURL *modelURL = [[NSBundle mainBundle] URLForResource:@"OEDatabase" withExtension:@"momd"];
    _managedObjectModel = [[NSManagedObjectModel alloc] initWithContentsOfURL:modelURL];

    return _managedObjectModel;
}

- (NSManagedObjectContext *)managedObjectContext
{
    return _managedObjectContext;
}

- (void)managedObjectContextDidSave:(NSNotification *)notification
{
    // This error checking is a bit redundant, but we want to make sure that we only merge in other thread's managed object contexts
    if([notification object] != _managedObjectContext)
    {
        //Transient properties don't merge
        [[notification userInfo][NSUpdatedObjectsKey] enumerateObjectsUsingBlock:^(id obj, NSUInteger idx, BOOL *stop) {
            if (![obj isKindOfClass:[OEDBGame class]])
                return;

            OEDBGame *incomingGame = obj;
            OEDBGame *mainGame = (OEDBGame *)[_managedObjectContext objectWithID:[incomingGame objectID]];
            [mainGame setStatus:[incomingGame status]];
        }];

        [_managedObjectContext performSelectorOnMainThread:@selector(mergeChangesFromContextDidSaveNotification:) withObject:notification waitUntilDone:YES];
    }
}

- (id)objectWithURI:(NSURL *)uri
{
    NSManagedObjectID *objID = [[self persistentStoreCoordinator] managedObjectIDForURIRepresentation:uri];
    return [[self managedObjectContext] objectWithID:objID];
}

#pragma mark -

- (BOOL)save:(NSError **)error
{
    if(![[self managedObjectContext] commitEditing])
    {
        NSLog(@"%@:%@ unable to commit editing before saving", [self class], NSStringFromSelector(_cmd));
        return NO;
    }

    if(![[self managedObjectContext] hasChanges])
    {
        NSLog(@"Database did not change. Skip Saving.");
        return YES;
    }

    if(![[self managedObjectContext] save:error])
    {
        [[NSApplication sharedApplication] presentError:*error];
        return NO;
    }

    return YES;
}

- (NSUndoManager *)undoManager
{
    return [[self managedObjectContext] undoManager];
}

#pragma mark - Database queries

/*
- (OEDBSystem *)systemWithIdentifier:(NSString *)identifier
{
    NSManagedObjectContext *context = [self managedObjectContext];
    NSEntityDescription *description = [OEDBSystem entityDescriptionInContext:context];
    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"systemIdentifier == %@", identifier];
    NSFetchRequest *fetchRequest = [[NSFetchRequest alloc] init];
    [fetchRequest setPredicate:predicate];
    [fetchRequest setFetchLimit:1];
    [fetchRequest setEntity:description];

    NSError *error = nil;

    NSArray *result = [context executeFetchRequest:fetchRequest error:&error];
    if(result == nil) NSLog(@"systemWithIdentifier: Error: %@", error);

    return [result lastObject];
}

- (OEDBSystem *)systemWithArchiveID:(NSNumber *)aID
{
    NSManagedObjectContext *context = [self managedObjectContext];
    NSEntityDescription *description = [OEDBSystem entityDescriptionInContext:context];
    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"archiveID == %ld", [aID integerValue]];
    NSFetchRequest *fetchRequest = [[NSFetchRequest alloc] init];
    [fetchRequest setPredicate:predicate];
    [fetchRequest setFetchLimit:1];
    [fetchRequest setEntity:description];

    NSError *error = nil;

    id result = [context executeFetchRequest:fetchRequest error:&error];
    if(!result)
    {
        NSLog(@"systemWithArchiveID: Error: %@", error);
        return nil;
    }
    return [result lastObject];
}
 */

- (NSUInteger)collectionsCount
{
    NSUInteger count = 1;
    NSManagedObjectContext *context = [self managedObjectContext];

    NSEntityDescription *descr = [NSEntityDescription entityForName:@"SmartCollection" inManagedObjectContext:context];
    NSFetchRequest *req = [[NSFetchRequest alloc] init];
    [req setEntity:descr];

    NSError *error = nil;
    NSUInteger ccount = [context countForFetchRequest:req error:&error];
    if(ccount == NSNotFound)
    {
        ccount = 0;
        NSLog(@"collectionsCount: Smart Collections Error: %@", error);
    }
    count += ccount;


    descr = [NSEntityDescription entityForName:@"Collection" inManagedObjectContext:context];
    req = [[NSFetchRequest alloc] init];
    [req setEntity:descr];

    ccount = [context countForFetchRequest:req error:&error];
    if(ccount == NSNotFound)
    {
        ccount = 0;
        NSLog(@"collectionsCount: Regular Collections Error: %@", error);
    }
    count += ccount;

    return count;
}

- (NSArray *)collections
{
    NSManagedObjectContext *context = [self managedObjectContext];
    NSMutableArray *collectionsArray = [NSMutableArray array];

    // insert "all games" item here !
    OEDBAllGamesCollection *allGamesCollections = [OEDBAllGamesCollection sharedDBAllGamesCollection];
    [collectionsArray addObject:allGamesCollections];

    NSEntityDescription *descr = [NSEntityDescription entityForName:@"SmartCollection" inManagedObjectContext:context];
    NSFetchRequest *req = [[NSFetchRequest alloc] init];

    NSSortDescriptor *sort = [[NSSortDescriptor alloc] initWithKey:@"name" ascending:YES selector:@selector(localizedStandardCompare:)];
    [req setSortDescriptors:[NSArray arrayWithObject:sort]];

    [req setEntity:descr];

    NSError *error = nil;

    id result = [context executeFetchRequest:req error:&error];
    if(result == nil)
    {
        NSLog(@"collections: Smart Collections Error: %@", error);
        return [NSArray array];
    }

    [collectionsArray addObjectsFromArray:result];

    descr = [NSEntityDescription entityForName:@"Collection" inManagedObjectContext:context];
    [req setEntity:descr];

    result = [context executeFetchRequest:req error:&error];
    if(result == nil)
    {
        NSLog(@"collections: Regular Collections Error: %@", error);
        return [NSArray array];
    }

    [collectionsArray addObjectsFromArray:result];

    return collectionsArray;
}

- (NSArray *)media
{
    NSMutableArray *mediaArray = [NSMutableArray array];
    // Saved Games
    OEDBSavedGamesMedia *savedGamesMedia = [OEDBSavedGamesMedia sharedDBSavedGamesMedia];
    [mediaArray addObject:savedGamesMedia];
    
    // Screenshots
    OEDBScreenshotsMedia *sreenshotsMedia = [OEDBScreenshotsMedia sharedDBScreenshotsMedia];
    [mediaArray addObject:sreenshotsMedia];
    
    // Video
    OEDBVideoMedia *videoMedia = [OEDBVideoMedia sharedDBVideoMedia];
    [mediaArray addObject:videoMedia];
    
    return mediaArray;
}

#pragma mark - Collection Editing

- (id)addNewCollection:(NSString *)name
{
    NSManagedObjectContext *context = [self managedObjectContext];

    if(name == nil)
    {
        name = NSLocalizedString(@"New Collection", @"Default collection name");

        NSEntityDescription *entityDescription = [NSEntityDescription entityForName:@"AbstractCollection" inManagedObjectContext:context];

        NSFetchRequest *request = [[NSFetchRequest alloc] init];
        [request setEntity:entityDescription];
        [request setFetchLimit:1];

        NSString *uniqueName = name;
        NSError *error = nil;
        int numberSuffix = 0;

        while([context countForFetchRequest:request error:&error] != 0 && error == nil)
        {
            numberSuffix++;
            uniqueName = [NSString stringWithFormat:@"%@ %d", name, numberSuffix];
            [request setPredicate:[NSPredicate predicateWithFormat:@"name == %@", uniqueName]];
        }

        name = uniqueName;
    }

    NSManagedObject *aCollection = [NSEntityDescription insertNewObjectForEntityForName:@"Collection" inManagedObjectContext:context];
    [aCollection setValue:name forKey:@"name"];

    return aCollection;
}

- (id)addNewSmartCollection:(NSString *)name
{
    NSManagedObjectContext *context = [self managedObjectContext];

    if(name == nil)
    {
        name = NSLocalizedString(@"New Smart Collection", @"");

        NSEntityDescription *entityDescription = [NSEntityDescription entityForName:@"AbstractCollection" inManagedObjectContext:context];

        NSFetchRequest *request = [[NSFetchRequest alloc] init];
        [request setEntity:entityDescription];
        [request setFetchLimit:1];

        NSString *uniqueName = name;
        NSError *error = nil;
        int numberSuffix = 0;

        while([context countForFetchRequest:request error:&error] != 0 && error == nil)
        {
            numberSuffix++;
            uniqueName = [NSString stringWithFormat:@"%@ %d", name, numberSuffix];
            [request setPredicate:[NSPredicate predicateWithFormat:@"name == %@", uniqueName]];
        }

        name = uniqueName;
    }

    NSManagedObject *aCollection = [NSEntityDescription insertNewObjectForEntityForName:@"SmartCollection" inManagedObjectContext:context];
    [aCollection setValue:name forKey:@"name"];

    return aCollection;
}

- (id)addNewCollectionFolder:(NSString *)name
{
    NSManagedObjectContext *context = [self managedObjectContext];

    if(name == nil)
    {
        name = NSLocalizedString(@"New Folder", @"");

        NSEntityDescription *entityDescription = [NSEntityDescription entityForName:@"AbstractCollection" inManagedObjectContext:context];

        NSFetchRequest *request = [[NSFetchRequest alloc] init];
        [request setEntity:entityDescription];
        [request setFetchLimit:1];

        NSString *uniqueName = name;
        NSError *error = nil;
        int numberSuffix = 0;

        while([context countForFetchRequest:request error:&error] != 0 && error == nil)
        {
            numberSuffix ++;
            uniqueName = [NSString stringWithFormat:@"%@ %d", name, numberSuffix];
            [request setPredicate:[NSPredicate predicateWithFormat:@"name == %@", uniqueName]];
        }
        name = uniqueName;
    }

    NSManagedObject *aCollection = [NSEntityDescription insertNewObjectForEntityForName:@"CollectionFolder" inManagedObjectContext:context];
    [aCollection setValue:name forKey:@"name"];

    return aCollection;
}


- (void)removeCollection:(NSManagedObject *)collection
{
    [[collection managedObjectContext] deleteObject:collection];
}

#pragma mark -

- (OEDBRom *)romForMD5Hash:(NSString *)hashString
{
    NSManagedObjectContext *context = [self managedObjectContext];
    NSEntityDescription *entityDescription = [NSEntityDescription entityForName:@"ROM" inManagedObjectContext:context];

    NSFetchRequest *fetchRequest = [[NSFetchRequest alloc] init];
    [fetchRequest setEntity:entityDescription];
    [fetchRequest setFetchLimit:1];
    [fetchRequest setIncludesPendingChanges:YES];

    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"md5 == %@", hashString];
    [fetchRequest setPredicate:predicate];

    NSError *err = nil;
    NSArray *result = [context executeFetchRequest:fetchRequest error:&err];
    if(result == nil)
    {
        NSLog(@"Error executing fetch request to get rom by md5");
        NSLog(@"%@", err);
        return nil;
    }

    return [result lastObject];
}

- (OEDBRom *)romForCRC32Hash:(NSString *)crc32String
{
    NSManagedObjectContext *context = [self managedObjectContext];
    NSEntityDescription *entityDescription = [NSEntityDescription entityForName:@"ROM" inManagedObjectContext:context];

    NSFetchRequest *fetchRequest = [[NSFetchRequest alloc] init];
    [fetchRequest setEntity:entityDescription];
    [fetchRequest setFetchLimit:1];
    [fetchRequest setIncludesPendingChanges:YES];

    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"crc32 == %@", crc32String];
    [fetchRequest setPredicate:predicate];

    NSError *err = nil;
    NSArray *result = [context executeFetchRequest:fetchRequest error:&err];
    if(result == nil)
    {
        NSLog(@"Error executing fetch request to get rom by crc");
        NSLog(@"%@", err);
        return nil;
    }

    return [result lastObject];
}

- (NSArray *)romsForPredicate:(NSPredicate *)predicate
{
    [_romsController setFilterPredicate:predicate];

    return [_romsController arrangedObjects];
}

- (NSArray *)romsInCollection:(id)collection
{
    // TODO: implement
    NSLog(@"Roms in collection called, but not implemented");
    return [NSArray array];
}

#pragma mark -

- (NSArray *)lastPlayedRoms
{
    // TODO: get numberOfRoms from defaults or system settings
    NSUInteger numberOfRoms = 5;
    NSFetchRequest *fetchRequest = [NSFetchRequest fetchRequestWithEntityName:[OEDBRom entityName]];

    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"lastPlayed != nil"];
    NSSortDescriptor *sortDesc = [NSSortDescriptor sortDescriptorWithKey:@"lastPlayed" ascending:NO];
    [fetchRequest setSortDescriptors:[NSArray arrayWithObject:sortDesc]];
    [fetchRequest setPredicate:predicate];
    [fetchRequest setFetchLimit:numberOfRoms];

    return [[self managedObjectContext] executeFetchRequest:fetchRequest error:nil];
}

- (NSDictionary *)lastPlayedRomsBySystem
{
    // TODO: get numberOfRoms from defaults or system settings
    NSUInteger numberOfRoms = 5;
    NSFetchRequest *fetchRequest = [NSFetchRequest fetchRequestWithEntityName:[OEDBRom entityName]];

    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"lastPlayed != nil"];
    NSSortDescriptor *sortDesc = [NSSortDescriptor sortDescriptorWithKey:@"lastPlayed" ascending:NO];
    [fetchRequest setSortDescriptors:[NSArray arrayWithObject:sortDesc]];
    [fetchRequest setPredicate:predicate];
    [fetchRequest setFetchLimit:numberOfRoms];

    NSArray *roms = [[self managedObjectContext] executeFetchRequest:fetchRequest error:nil];
    NSMutableSet *systemsSet = [NSMutableSet setWithCapacity:[roms count]];
    [roms enumerateObjectsUsingBlock:
     ^(OEDBRom *aRom, NSUInteger idx, BOOL *stop)
     {
         [systemsSet addObject:[[aRom game] system]];
     }];

    NSArray *systems = [systemsSet allObjects];
    NSMutableDictionary *result = [NSMutableDictionary dictionaryWithCapacity:[systems count]];
    [systems enumerateObjectsUsingBlock:
     ^(OEDBSystem *aSystem, NSUInteger idx, BOOL *stop)
     {
         NSArray *romsForSystem = [roms filteredArrayUsingPredicate:
                                   [NSPredicate predicateWithBlock:
                                    ^ BOOL (id aRom, NSDictionary *bindings)
                                    {
                                        return [[aRom game] system] == aSystem;
                                    }]];
         [result setObject:romsForSystem forKey:[aSystem lastLocalizedName]];
     }];

    return result;
}

#pragma mark - Datbase Folders

- (NSURL *)databaseFolderURL
{
    NSUserDefaults *standardDefaults = [NSUserDefaults standardUserDefaults];
    NSString *libraryFolderPath = [standardDefaults stringForKey:OEDatabasePathKey];

    return [NSURL fileURLWithPath:libraryFolderPath isDirectory:YES];
}

- (NSURL *)romsFolderURL
{
    NSPersistentStore *persistentStore = [[[self persistentStoreCoordinator] persistentStores] lastObject];
    NSDictionary *metadata = [[self persistentStoreCoordinator] metadataForPersistentStore:persistentStore];
    if([metadata objectForKey:OELibraryRomsFolderURLKey])
    {
        NSString *urlString = [metadata objectForKey:OELibraryRomsFolderURLKey];
        if([urlString rangeOfString:@"file://"].location == NSNotFound)
            return [NSURL URLWithString:urlString relativeToURL:[self databaseFolderURL]];
        return [NSURL URLWithString:urlString];
    }
    else
    {
        NSURL *result = [[self databaseFolderURL] URLByAppendingPathComponent:@"roms" isDirectory:YES];
        [[NSFileManager defaultManager] createDirectoryAtURL:result withIntermediateDirectories:YES attributes:nil error:nil];
        [self setRomsFolderURL:result];
        return result;
    }
}

- (void)setRomsFolderURL:(NSURL *)url
{
    if(url != nil)
    {
        NSError             *error             = nil;
        NSPersistentStore   *persistentStore   = [[[self persistentStoreCoordinator] persistentStores] lastObject];
        NSDictionary        *metadata          = [persistentStore metadata];
        NSMutableDictionary *mutableMetaData   = [metadata mutableCopy];
        NSURL               *databaseFolderURL = [self databaseFolderURL];

        if([url isSubpathOfURL:databaseFolderURL])
        {
            NSString *urlString = [[url absoluteString] substringFromIndex:[[databaseFolderURL absoluteString] length]];
            [mutableMetaData setObject:[@"./" stringByAppendingString:urlString] forKey:OELibraryRomsFolderURLKey];
        }
        else [mutableMetaData setObject:[url absoluteString] forKey:OELibraryRomsFolderURLKey];

        // Using the instance method sets the metadata for the current store in memory, while
        // using the class method writes to disk immediately. Calling both seems redundant
        // but is the only way i found that works.
        //
        // Also see discussion at http://www.cocoabuilder.com/archive/cocoa/295041-setting-not-saving-nspersistentdocument-metadata-changes-file-modification-date.html
        [[self persistentStoreCoordinator] setMetadata:mutableMetaData forPersistentStore:persistentStore];
        [NSPersistentStoreCoordinator setMetadata:mutableMetaData forPersistentStoreOfType:[persistentStore type] URL:[persistentStore URL] error:&error];

        [self save:nil];
    }
}

- (NSURL *)unsortedRomsFolderURL
{
    NSString *unsortedFolderName = NSLocalizedString(@"unsorted", @"Unsorted Folder Name");

    NSURL *result = [[self romsFolderURL] URLByAppendingPathComponent:unsortedFolderName isDirectory:YES];
    [[NSFileManager defaultManager] createDirectoryAtURL:result withIntermediateDirectories:YES attributes:nil error:nil];

    return result;
}

- (NSURL *)romsFolderURLForSystem:(OEDBSystem *)system
{
    if([system name] == nil) return [self unsortedRomsFolderURL];

    NSURL *result = [[self romsFolderURL] URLByAppendingPathComponent:[system name] isDirectory:YES];
    [[NSFileManager defaultManager] createDirectoryAtURL:result withIntermediateDirectories:YES attributes:nil error:nil];

    return result;
}

- (NSURL *)stateFolderURL
{
    if([[NSUserDefaults standardUserDefaults] objectForKey:OESaveStateFolderURLKey]) return [NSURL URLWithString:[[NSUserDefaults standardUserDefaults] objectForKey:OESaveStateFolderURLKey]];

    NSString *saveStateFolderName = NSLocalizedString(@"Save States", @"Save States Folder Name");
    NSURL    *result = [[NSFileManager defaultManager] URLForDirectory:NSApplicationSupportDirectory inDomain:NSUserDomainMask appropriateForURL:nil create:NO error:nil];
    result = [result URLByAppendingPathComponent:@"OpenEmu" isDirectory:YES];
    result = [result URLByAppendingPathComponent:saveStateFolderName isDirectory:YES];

    // [[NSFileManager defaultManager] createDirectoryAtURL:result withIntermediateDirectories:YES attributes:nil error:nil];

    return result;
}

- (NSURL *)stateFolderURLForSystem:(OEDBSystem *)system
{
    OESystemPlugin *plugin = [system plugin];
    NSString *displayName = [plugin displayName] ?: NSLocalizedString(@"Unkown System", "Name of directory for savestates with unknown systems");
    NSURL *result = [[self stateFolderURL] URLByAppendingPathComponent:displayName isDirectory:YES];
    [[NSFileManager defaultManager] createDirectoryAtURL:result withIntermediateDirectories:YES attributes:nil error:nil];

    return result;
}

- (NSURL *)stateFolderURLForROM:(OEDBRom *)rom
{
    NSURL *result = [[self stateFolderURLForSystem:[[rom game] system]] URLByAppendingPathComponent:[[[rom URL] lastPathComponent] stringByDeletingPathExtension]];
    [[NSFileManager defaultManager] createDirectoryAtURL:result withIntermediateDirectories:YES attributes:nil error:nil];
    return result;
}

- (NSURL *)screenshotFolderURL
{
    if([[NSUserDefaults standardUserDefaults] objectForKey:OEScreenshotFolderURLKey])
        return [NSURL URLWithString:[[NSUserDefaults standardUserDefaults] objectForKey:OEScreenshotFolderURLKey]];

    NSString *screenshotFolderName = NSLocalizedString(@"Screenshots", @"Screenshot Folder Name");
    NSURL    *result = [[NSFileManager defaultManager] URLForDirectory:NSApplicationSupportDirectory inDomain:NSUserDomainMask appropriateForURL:nil create:NO error:nil];
    result = [result URLByAppendingPathComponent:@"OpenEmu" isDirectory:YES];
    result = [result URLByAppendingPathComponent:screenshotFolderName isDirectory:YES];

    [[NSFileManager defaultManager] createDirectoryAtURL:result withIntermediateDirectories:YES attributes:nil error:nil];

    return result;
}

- (NSURL *)coverFolderURL
{
    NSUserDefaults *standardDefaults  = [NSUserDefaults standardUserDefaults];
    NSString       *libraryFolderPath = [standardDefaults stringForKey:OEDatabasePathKey];
    NSString       *coverFolderPath   = [libraryFolderPath stringByAppendingPathComponent:@"Artwork/"];

    NSURL *url = [NSURL fileURLWithPath:coverFolderPath isDirectory:YES];
    [[NSFileManager defaultManager] createDirectoryAtURL:url withIntermediateDirectories:YES attributes:nil error:nil];
    return url;
}

- (NSURL *)importQueueURL
{
    NSURL *baseURL = [self databaseFolderURL];
    return [baseURL URLByAppendingPathComponent:@"Import Queue.db"];
}

#pragma mark - ArchiveVG Sync

- (void)startArchiveVGSync
{
    if(_syncThread == nil || [_syncThread isFinished])
    {
        _syncThread = [[NSThread alloc] initWithTarget:self selector:@selector(archiveVGSyncThread) object:nil];
        [_syncThread start];
    }
}

- (void)archiveVGSyncThread
{
    NSArray        *result    = nil;
    NSError        *error     = nil;
    NSFetchRequest *request   = [[NSFetchRequest alloc] initWithEntityName:[OEDBGame entityName]];
    NSPredicate    *predicate = [NSPredicate predicateWithFormat:@"status == %d", OEDBGameStatusProcessing];

    [request setFetchLimit:1];
    [request setPredicate:predicate];

    while((result = [[self managedObjectContext] executeFetchRequest:request error:&error]) && [result count])
    {
        @autoreleasepool
        {
            OEDBGame *game = [result lastObject];
            [game performArchiveSync];
        }
    }
}

#pragma mark - Debug

- (void)dump
{
    [self dumpWithPrefix:@"***"];
}

- (void)dumpWithPrefix:(NSString *)prefix
{
    NSString *subPrefix = [prefix stringByAppendingString:@"-----"];
    NSLog(@"%@ Beginning of database dump\n", prefix);

    NSLog(@"%@ Database folder is %@", prefix, [[self databaseFolderURL] path]);
    NSLog(@"%@ Number of collections is %lu", prefix, (unsigned long)[self collectionsCount]);

    for(id collection in [self collections])
    {
        if([collection respondsToSelector:@selector(dumpWithPrefix:)]) [collection dumpWithPrefix:subPrefix];
        else NSLog(@"%@ Collection is %@", subPrefix, collection);
    }

    NSLog(@"%@", prefix);
    NSLog(@"%@ Number of systems is %lu", prefix, (unsigned long)[OEDBSystem systemsCountInDatabase:self]);
    for(id system in [OEDBSystem allSystemsInDatabase:self])
    {
        if([system respondsToSelector:@selector(dumpWithPrefix:)]) [system dumpWithPrefix:subPrefix];
        else NSLog(@"%@ System is %@", subPrefix, system);
    }

    NSLog(@"%@", prefix);
    NSLog(@"%@ ALL ROMs", prefix);
    for(id ROM in [self allROMsForDump])
    {
        if([ROM respondsToSelector:@selector(dumpWithPrefix:)]) [ROM dumpWithPrefix:subPrefix];
        else NSLog(@"%@ ROM is %@", subPrefix, ROM);
    }

    NSLog(@"%@ End of database dump\n\n", prefix);
}

- (NSArray *)allROMsForDump
{
    NSManagedObjectContext *moc = [self managedObjectContext];
    NSEntityDescription *entity = [NSEntityDescription entityForName:@"ROM" inManagedObjectContext:moc];
    NSFetchRequest *fetchReq = [NSFetchRequest new];
    [fetchReq setEntity:entity];
    return [moc executeFetchRequest:fetchReq error:NULL];
}

@end
